@@ -122,10 +122,10 @@ func arrValues(words ...*Word) *ArrayExpr {
 
 type testCase struct {
 	Strs        []string
-	common      interface{}
-	bash, posix interface{}
-	bsmk, mksh  interface{}
-	bats        interface{}
+	common      any
+	bash, posix any
+	bsmk, mksh  any
+	bats        any
 	All         []*File
 	Bash, Posix *File
 	MirBSDKorn  *File
@@ -1546,15 +1546,6 @@ var fileTests = []testCase{
 				{Op: AppAll, Word: litWord("b")},
 			},
 		},
-		posix: []*Stmt{
-			{Cmd: litCall("foo"), Background: true},
-			{Redirs: []*Redirect{
-				{Op: RdrOut, Word: litWord("a")},
-			}, Background: true},
-			{Redirs: []*Redirect{
-				{Op: AppOut, Word: litWord("b")},
-			}},
-		},
 	},
 	{
 		Strs: []string{"foo 2>file bar", "2>file foo bar"},
@@ -1743,6 +1734,13 @@ var fileTests = []testCase{
 		Strs:   []string{"foo#bar"},
 		common: litWord("foo#bar"),
 	},
+	{
+		Strs: []string{"$foo#bar foo#$bar"},
+		common: call(
+			word(litParamExp("foo"), lit("#bar")),
+			word(lit("foo#"), litParamExp("bar")),
+		),
+	},
 	{
 		Strs:   []string{"{ echo } }; }"},
 		common: block(litStmt("echo", "}", "}")),
@@ -1799,7 +1797,7 @@ var fileTests = []testCase{
 	{
 		Strs: []string{
 			`$(echo '\' 'a\b' "\\" "a\a")`,
-			"`" + `echo '\' 'a\b' "\\\\" "a\a"` + "`",
+			"`" + `echo '\' 'a\\b' "\\\\" "a\a"` + "`",
 		},
 		common: cmdSubst(stmt(call(
 			litWord("echo"),
@@ -4518,7 +4516,7 @@ var fileTestsKeepComments = []testCase{
 	},
 }
 
-func fullProg(v interface{}) *File {
+func fullProg(v any) *File {
 	f := &File{}
 	switch x := v.(type) {
 	case *File:
@@ -4547,7 +4545,7 @@ func fullProg(v interface{}) *File {
 	return nil
 }
 
-func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
+func recursiveSanityCheck(tb testing.TB, src string, v any) {
 	checkPos := func(pos Pos, strs ...string) {
 		if !pos.IsValid() {
 			tb.Fatalf("invalid Pos in %T", v)
@@ -4591,7 +4589,7 @@ func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
 				return
 			}
 		}
-		tb.Errorf("Expected one of %q at %d in %q, found %q",
+		tb.Errorf("Expected one of %q at %s in %q, found %q",
 			strs, pos, src, gotErr)
 	}
 	checkNodePosEnd := func(n Node) {
@@ -4599,7 +4597,7 @@ func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
 			tb.Errorf("Found End() before Pos() in %T", n)
 		}
 	}
-	recurse := func(v interface{}) {
+	recurse := func(v any) {
 		recursiveSanityCheck(tb, src, v)
 		if n, ok := v.(Node); ok {
 			checkNodePosEnd(n)
@@ -4633,6 +4631,8 @@ func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
 				// ended by semicolon
 			case endOff > 0 && src[endOff-1] == '&':
 				// ended by & or |&
+			case end == '\\' && src[endOff+1] == '`':
+				// ended by an escaped backquote
 			default:
 				tb.Errorf("Unexpected Stmt.End() %d %q in %q",
 					endOff, end, src)
@@ -4701,7 +4701,7 @@ func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
 			tb.Errorf("Lit without newlines has Pos/End lines %d and %d",
 				posLine, endLine)
 		case strings.Contains(src, "`") && strings.Contains(src, "\\"):
-			// removed quotes inside backquote cmd substs
+			// removed backslashes inside backquote cmd substs
 			val = ""
 		case end < len(src) && (src[end] == '\n' || src[end] == '`'):
 			// heredoc literals that end with the
@@ -4794,7 +4794,12 @@ func recursiveSanityCheck(tb testing.TB, src string, v interface{}) {
 		if x.Dollar {
 			valuePos = posAddCol(valuePos, 1)
 		}
-		checkPos(valuePos, x.Value)
+		val := x.Value
+		if strings.Contains(src, "`") && strings.Contains(src, "\\") {
+			// removed backslashes inside backquote cmd substs
+			val = ""
+		}
+		checkPos(valuePos, val)
 		if x.Dollar {
 			checkPos(x.Left, "$'")
 		} else {
