@@ -15,6 +15,7 @@ import (
 	"sync"
 	"testing"
 
+	qt "github.com/frankban/quicktest"
 	"github.com/google/go-cmp/cmp"
 )
 
@@ -155,61 +156,44 @@ func TestParseBats(t *testing.T) {
 }
 
 func TestMain(m *testing.M) {
-	// Set the locale to computer-friendly English and UTF-8.
-	// Some systems like Arch miss C.UTF8, so fall back to the US English locale.
-	//
-	// TODO: glibc 2.35 was released with C.UTF-8 in February 2022.
-	// This locale is now becoming a standard, so remove the workaround in 2023.
-	if out, _ := exec.Command("locale", "-a").Output(); strings.Contains(
-		strings.ToLower(string(out)), "c.utf",
-	) {
-		os.Setenv("LANGUAGE", "C.UTF-8")
-		os.Setenv("LC_ALL", "C.UTF-8")
-	} else {
-		os.Setenv("LANGUAGE", "en_US.UTF-8")
-		os.Setenv("LC_ALL", "en_US.UTF-8")
-	}
+	// Set the locale to computer-friendly English and UTF-8, C.UTF-8,
+	// which started shipping with glibc 2.35 in February 2022.
+	os.Setenv("LANGUAGE", "C.UTF-8")
+	os.Setenv("LC_ALL", "C.UTF-8")
 	os.Exit(m.Run())
 }
 
 var (
-	storedHasBash51 bool
-	onceHasBash51   sync.Once
+	onceHasBash52 = sync.OnceValue(func() bool {
+		return cmdContains("version 5.2", "bash", "--version")
+	})
 
-	storedHasDash059 bool
-	onceHasDash059   sync.Once
+	onceHasDash059 = sync.OnceValue(func() bool {
+		// dash provides no way to check its version, so we have to
+		// check if it's new enough as to not have the bug that breaks
+		// our integration tests.
+		// This also means our check does not require a specific version.
+		return cmdContains("Bad subst", "dash", "-c", "echo ${#<}")
+	})
 
-	storedHasMksh59 bool
-	onceHasMksh59   sync.Once
+	onceHasMksh59 = sync.OnceValue(func() bool {
+		return cmdContains(" R59 ", "mksh", "-c", "echo $KSH_VERSION")
+	})
 )
 
-func hasBash51(tb testing.TB) {
-	onceHasBash51.Do(func() {
-		storedHasBash51 = cmdContains("version 5.1", "bash", "--version")
-	})
-	if !storedHasBash51 {
-		tb.Skipf("bash 5.1 required to run")
+func requireBash52(tb testing.TB) {
+	if !onceHasBash52() {
+		tb.Skipf("bash 5.2 required to run")
 	}
 }
-
-func hasDash059(tb testing.TB) {
-	// dash provides no way to check its version, so we have to
-	// check if it's new enough as to not have the bug that breaks
-	// our integration tests.
-	// This also means our check does not require a specific version.
-	onceHasDash059.Do(func() {
-		storedHasDash059 = cmdContains("Bad subst", "dash", "-c", "echo ${#<}")
-	})
-	if !storedHasDash059 {
+func requireDash059(tb testing.TB) {
+	if !onceHasDash059() {
 		tb.Skipf("dash 0.5.9+ required to run")
 	}
 }
 
-func hasMksh59(tb testing.TB) {
-	onceHasMksh59.Do(func() {
-		storedHasMksh59 = cmdContains(" R59 ", "mksh", "-c", "echo $KSH_VERSION")
-	})
-	if !storedHasMksh59 {
+func requireMksh59(tb testing.TB) {
+	if !onceHasMksh59() {
 		tb.Skipf("mksh 59 required to run")
 	}
 }
@@ -242,13 +226,13 @@ func confirmParse(in, cmd string, wantErr bool) func(*testing.T) {
 			// -n makes bash accept invalid inputs like
 			// "let" or "`{`", so only use it in
 			// non-erroring tests. Should be safe to not use
-			// -n anyway since these are supposed to just
-			// fail.
+			// -n anyway since these are supposed to just fail.
 			// also, -n will break if we are using extglob
 			// as extglob is not actually applied.
 			opts = append(opts, "-n")
 		}
 		cmd := exec.Command(cmd, opts...)
+		cmd.Dir = t.TempDir() // to be safe
 		cmd.Stdin = strings.NewReader(in)
 		var stderr bytes.Buffer
 		cmd.Stderr = &stderr
@@ -274,7 +258,7 @@ func TestParseBashConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling bash is slow.")
 	}
-	hasBash51(t)
+	requireBash52(t)
 	i := 0
 	for _, c := range append(fileTests, fileTestsNoPrint...) {
 		if c.Bash == nil {
@@ -292,7 +276,7 @@ func TestParsePosixConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling dash is slow.")
 	}
-	hasDash059(t)
+	requireDash059(t)
 	i := 0
 	for _, c := range append(fileTests, fileTestsNoPrint...) {
 		if c.Posix == nil {
@@ -310,7 +294,7 @@ func TestParseMirBSDKornConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling mksh is slow.")
 	}
-	hasMksh59(t)
+	requireMksh59(t)
 	i := 0
 	for _, c := range append(fileTests, fileTestsNoPrint...) {
 		if c.MirBSDKorn == nil {
@@ -328,7 +312,7 @@ func TestParseErrBashConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling bash is slow.")
 	}
-	hasBash51(t)
+	requireBash52(t)
 	for _, c := range shellTests {
 		want := c.common
 		if c.bsmk != nil {
@@ -349,7 +333,7 @@ func TestParseErrPosixConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling dash is slow.")
 	}
-	hasDash059(t)
+	requireDash059(t)
 	for _, c := range shellTests {
 		want := c.common
 		if c.posix != nil {
@@ -367,7 +351,7 @@ func TestParseErrMirBSDKornConfirm(t *testing.T) {
 	if testing.Short() {
 		t.Skip("calling mksh is slow.")
 	}
-	hasMksh59(t)
+	requireMksh59(t)
 	for _, c := range shellTests {
 		want := c.common
 		if c.bsmk != nil {
@@ -429,9 +413,9 @@ func BenchmarkParse(b *testing.B) {
 
 type errorCase struct {
 	in          string
-	common      interface{}
-	bash, posix interface{}
-	bsmk, mksh  interface{}
+	common      any
+	bash, posix any
+	bsmk, mksh  any
 }
 
 var shellTests = []errorCase{
@@ -811,6 +795,14 @@ var shellTests = []errorCase{
 		in:     "echo && > #",
 		common: `1:9: > must be followed by a word`,
 	},
+	{
+		in:    "foo &>/dev/null",
+		posix: `1:5: &> redirects are a bash/mksh feature`,
+	},
+	{
+		in:    "foo &>>/dev/null",
+		posix: `1:5: &> redirects are a bash/mksh feature`,
+	},
 	{
 		in:     "<<",
 		common: `1:1: << must be followed by a word`,
@@ -1858,15 +1850,19 @@ var shellTests = []errorCase{
 	},
 	{
 		in:    "function foo() { bar; }",
-		posix: `1:13: the "function" builtin exists in bash; tried parsing as posix`,
+		posix: `1:13: the "function" builtin is a bash feature; tried parsing as posix`,
+	},
+	{
+		in:    "function foo { bar; }",
+		posix: `1:14: the "function" builtin is a bash feature; tried parsing as posix`,
 	},
 	{
 		in:    "declare foo=(bar)",
-		posix: `1:13: the "declare" builtin exists in bash; tried parsing as posix`,
+		posix: `1:13: the "declare" builtin is a bash feature; tried parsing as posix`,
 	},
 	{
 		in:    "let foo=(bar)",
-		posix: `1:9: the "let" builtin exists in bash; tried parsing as posix`,
+		posix: `1:9: the "let" builtin is a bash feature; tried parsing as posix`,
 	},
 	{
 		in:    "echo <(",
@@ -2418,7 +2414,7 @@ func TestParseArithmeticError(t *testing.T) {
 var stopAtTests = []struct {
 	in   string
 	stop string
-	want interface{}
+	want any
 }{
 	{
 		"foo bar", "$$",
@@ -2527,3 +2523,20 @@ func TestIsIncomplete(t *testing.T) {
 		})
 	}
 }
+
+func TestBackquotesPos(t *testing.T) {
+	in := "`\\\\foo`"
+	p := NewParser()
+	f, err := p.Parse(strings.NewReader(in), "")
+	qt.Assert(t, err, qt.IsNil)
+	cmdSubst := f.Stmts[0].Cmd.(*CallExpr).Args[0].Parts[0].(*CmdSubst)
+	lit := cmdSubst.Stmts[0].Cmd.(*CallExpr).Args[0].Parts[0].(*Lit)
+
+	qt.Assert(t, lit.Value, qt.Equals, lit.Value)
+	// Note that positions of literals with escape sequences inside backquote command substitutions
+	// are weird, since we effectively skip over the double escaping in the literal value and positions.
+	// Even though the input source has '\\foo' between columns 2 and 7 (length 5)
+	// we end up keeping '\foo' between columns 3 and 7 (length 4).
+	qt.Assert(t, lit.ValuePos.String(), qt.Equals, "1:2")
+	qt.Assert(t, lit.ValueEnd.String(), qt.Equals, "1:7")
+}
