@@ -91,6 +91,9 @@ var printTests = []printCase{
 	{"if a; then b\nelse c\nfi", "if a; then\n\tb\nelse\n\tc\nfi"},
 	samePrint("foo >&2 <f bar"),
 	samePrint("foo >&2 bar <f"),
+	samePrint(">&2 foo bar <f"),
+	samePrint(">&2 foo"),
+	samePrint(">&2 foo 2>&1 bar <f"),
 	{"foo >&2>/dev/null", "foo >&2 >/dev/null"},
 	{"foo <<EOF bar\nl1\nEOF", "foo bar <<EOF\nl1\nEOF"},
 	samePrint("foo <<\\\\\\\\EOF\nbar\n\\\\EOF"),
@@ -620,7 +623,7 @@ var printTests = []printCase{
 	samePrint("#comment\n>redir"),
 	{
 		">redir \\\n\tfoo",
-		"foo >redir",
+		">redir foo",
 	},
 	samePrint("$(declare)"),
 	{
@@ -649,6 +652,13 @@ var printTests = []printCase{
 	},
 	samePrint("(\n\t((foo++))\n)"),
 	samePrint("(foo && bar)"),
+	samePrint(`$foo#bar ${foo}#bar 'foo'#bar "foo"#bar`),
+	// TODO: support cases with command substitutions as well
+	// {
+	// 	"`foo`#bar",
+	// 	"$(foo)#bar",
+	// },
+	// samePrint(`$("foo"#bar)#bar`),
 }
 
 func TestPrintWeirdFormat(t *testing.T) {
@@ -1068,6 +1078,15 @@ func TestPrintMinify(t *testing.T) {
 		},
 		samePrint("foo >bar 2>baz <etc"),
 		samePrint("<<-EOF\n$(a|b)\nEOF"),
+		{
+			"a=$(\n\tcat <<'EOF'\n  hello\nEOF\n)",
+			"a=$(cat <<'EOF'\n  hello\nEOF\n)",
+		},
+		{
+			"(\n\tcat <<EOF\n hello\nEOF\n)",
+			"(cat <<EOF\n hello\nEOF\n)",
+		},
+		samePrint("diff -y <(cat <<EOF\n1\n2\n3\nEOF\n) <(cat <<EOF\n1\n4\n3\nEOF\n)"),
 	}
 	parser := NewParser(KeepComments(true))
 	printer := NewPrinter(Minify(true))
@@ -1161,8 +1180,8 @@ func TestPrintOptionsNotBroken(t *testing.T) {
 		{"SingleLine", []PrinterOption{SingleLine(true)}},
 	} {
 		printer := NewPrinter(opts.list...)
-		for i, tc := range append(fileTests, fileTestsNoPrint...) {
-			t.Run(fmt.Sprintf("File%s%03d", opts.name, i), func(t *testing.T) {
+		for _, tc := range append(fileTests, fileTestsNoPrint...) {
+			t.Run("File"+opts.name, func(t *testing.T) {
 				parser := parserPosix
 				if tc.Bats != nil {
 					parser = parserBats
@@ -1189,8 +1208,8 @@ func TestPrintOptionsNotBroken(t *testing.T) {
 				}
 			})
 		}
-		for i, tc := range printTests {
-			t.Run(fmt.Sprintf("Print%s%03d", opts.name, i), func(t *testing.T) {
+		for _, tc := range printTests {
+			t.Run("Print"+opts.name, func(t *testing.T) {
 				prog, err := parserBash.Parse(strings.NewReader(tc.in), "")
 				if err != nil {
 					t.Fatal(err)
